
/* 
   Copyright 2006, Christian Pellegrin <chri@evolware.com> and  ATLab

   This software is relased under the GPL (GNU Public License) v2.
   This is free software with absolutely no warranty

   TODO: check locking requirements, now we are not SMP aware (need to
   protect tasklet from itself running on other processors, otherwise
   it should be allright)


   version 0.1: first public release

 */



//#define DEBUG 1


//#define SHOWP 1 /* define this to dump packets */

#ifdef SHOWP
#define DUMPP(P) \
 { \
   int i; \
   struct sk_buff *s = P; \
   unsigned char *d = s->data; \
   printk("%d: ", s->len); \
   for(i=0; i<s->len; i++) { \
     printk("%02x ", d[i]); \
   } \
   printk("\n"); \
 }
#else
#define DUMPP(P) 
#endif

#if 0				/* define this to see start/stop of iface */
#define STARTI  printk("start\n")
#define STOPI   printk("stop\n")
#define TXI     printk("txe\n")
#define TXA     printk("txa\n")
#define RXI(X)     printk("rxe %d\n", X)
#define TXS(X)     printk("txs %d\n", X)
#define TXT(X)     printk("txt %d\n", X)
#define RXS     printk("rxs\n")
#define WW(S)      printk("ww %s\n", S)
#define WB(X)      printk("wb %02x\n", X)
#else
#define STARTI  
#define STOPI   
#define TXI
#define TXA
#define RXI(X)
#define TXS(X)
#define TXT(X)     
#define RXS
#define WW(S)
#define WB(X)
#endif

#ifdef DEBUG
#define PRREG(ch, x)   pr_debug(#x " %02x\n", serocco_readb(ch, x))
#endif

/* define this if you want to see *all* io */
#define DEBUG_IO 1

/* debug full status */
//#define DEBUG_FULL 1

#include <linux/config.h>

#include <linux/module.h>
#include <linux/ioport.h>
#include <linux/init.h>
#include <linux/console.h>
#include <linux/sysrq.h>
#include <linux/serial_reg.h>
#include <linux/circ_buf.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/device.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial_core.h>

#include <asm/io.h>
#include <asm/hardware.h>
#include <asm/irq.h>
#include <asm/arch/pxa-regs.h>

#include "serocco.h"

/* HW specific include */
#include <asm/arch/amesom.h>

/* configurable defines */

#define SEROCCO_BUS_MULT 4

#define SEROCCO_RESET_PIN 26

#define SEROCCO_IRQ_PIN 22

/* clock for the serocco */
#define UART_CLOCK 14745600

/* define this to 3 to force loopback otherwise 0 */
//#define FORCE_LOOPBACK 3
#define FORCE_LOOPBACK 0

/* define this tou use async timeout for bigger fifo */
#define BIGTO 1

/* TODO: check this */
#define SEROCCO_MTU			2047

/* TODO, study this */ 
//#define SEROCCO_ADDRESS_MODE 1

#define SEROCCO_HDLC_ADDRESS 0x0069

#if SEROCCO_ADDRESS_MODE == 1
int address = SEROCCO_HDLC_ADDRESS;
MODULE_PARM(address, "i");
MODULE_PARM_DESC(address, "HDLC address (default " \
                          __MODULE_STRING(SEROCCO_HDLC_ADDRESS) ")");
#endif


/* global pointer to chip structure */
/* TODO: get rid of this so we can have many instances */

static struct serocco_chip * my_serocco; /* global data structure */


int serocco_port_a_mode = SEROCCO_MODE_ASYNC; /* 3 */
MODULE_PARM(serocco_port_a_mode, "i");
MODULE_PARM_DESC(serocco_port_a_mode, "port a mode, 1 async, 2 sync");

/* this forces clock to given mode (for testing) */
static int force_clock = -1;
MODULE_PARM(force_clock, "i");
MODULE_PARM_DESC(force_clock, "force clock to given mode (TOE, SSEL, CM (5 bits) see manual) ONLY ASYNC ONLY TESTING");

/* this forces clock to master/slave */
static int force_master = 0;
MODULE_PARM(force_master, "i");
MODULE_PARM_DESC(force_master, "force master mode");

/* this forces baud to master */
static int force_baud = 64000;
MODULE_PARM(force_baud, "i");
MODULE_PARM_DESC(force_baud, "force baud");

int serocco_port_b_mode = SEROCCO_MODE_OFF; /* -1 */
MODULE_PARM(serocco_port_b_mode, "i");
MODULE_PARM_DESC(serocco_port_b_mode, "port a mode, 1 async, 2 sync");


/* HW specific start */
/* TODO: move this in platform code */

struct serocco_pin_s serocco_pins_port_a[] =
  {
    { TIOCM_DSR, SEROCCO_GPDATH, (1<<0), 1, SEROCCO_GPIMH, (1<<0)},
    { TIOCM_DTR, SEROCCO_GPDATH, (1<<1), 1, 0, 0},
    { TIOCM_RI,  SEROCCO_GPDATH, (1<<2), 1, SEROCCO_GPIMH, (1<<2)},
    { TIOCM_CTS, SEROCCO_STARL, (1<<0), 0, SEROCCO_IMR1, SEROCCO_CSC},
    { TIOCM_RTS, SEROCCO_CCR1H, SEROCCO_FRTS, 1, 0, 0},
    { TIOCM_CAR, SEROCCO_STARH, SEROCCO_CD, 1, SEROCCO_IMR2, SEROCCO_CDSC},
    { 0,},
  };

struct serocco_pin_s serocco_pins_port_b[] =
  {
    { TIOCM_CTS, SEROCCO_STARL, (1<<0), 0, SEROCCO_IMR1, SEROCCO_CSC},
    { TIOCM_RTS, SEROCCO_CCR1H, SEROCCO_FRTS, 1, 0, 0},
    { TIOCM_CAR, SEROCCO_STARH, SEROCCO_CD, 1, SEROCCO_IMR2, SEROCCO_CDSC},
    { 0,},
  };

/* shouldn't be more configurable parameters after this .... */

/* generic i/o handling, note register shift is HW specific */

inline static void serocco_writeb(struct serocco_ch *ch, u8 b, u32 reg) {
  u32 from;
  
  if (reg >= 0xb0 && reg <= 0xc9) {
    if (ch->off)
      from = ch->chip->base + ((reg + 0x1a)*SEROCCO_BUS_MULT);
    else
      from = ch->chip->base + (reg*SEROCCO_BUS_MULT);    
  }
  else {
    if (reg < 0x10 || reg > 0xe4) {
      from = ch->chip->base + (reg*SEROCCO_BUS_MULT);
    }
    else {
      from = ch->chip->base + ((reg + ch->off)*SEROCCO_BUS_MULT);
    }
  }
  writel(b, from);
#ifdef DEBUG_IO
  pr_debug("w %02x (%08x) = %02x\n", reg, from, b);
#endif
}

inline static u8 serocco_readb(struct serocco_ch *ch, u32 reg) {
  u32 r;
  u32 from;

  if (reg >= 0xb0 && reg <= 0xc9) {
    if (ch->off)
      from = ch->chip->base + ((reg + 0x1a)*SEROCCO_BUS_MULT);
    else
      from = ch->chip->base + (reg*SEROCCO_BUS_MULT);    
  }
  else {
    if (reg < 0x10 || reg > 0xe4) {
      from = ch->chip->base + (reg*SEROCCO_BUS_MULT);
    }
    else {
      from = ch->chip->base + ((reg + ch->off)*SEROCCO_BUS_MULT);
    }
  }
  r = readl(from);
#ifdef DEBUG_IO
  pr_debug("r %02x (%08x) = %02x\n", reg, from, r);
#endif
  return r & 0xff;
}

#ifdef DEBUG_FULL
void serocco_full_dump(struct serocco_ch *ch) {
  int i;
  u32 r, from;

  for(i=0; i<256; i++) {
    from = ch->chip->base + (i*4);
    r = readl(from);
    printk("%02x ", r & 0xff);
    if (i % 16 == 15) 
      printk("\n");
  }
}
#endif

static void serocco_get(struct serocco_ch *ch) {
  if (!in_atomic()) {
    tasklet_disable(&ch->chip->tasklet);
    down(&ch->chip->lock);
  }
}

static void serocco_put(struct serocco_ch *ch) {
  if (!in_atomic()) {
    up(&ch->chip->lock);
    tasklet_enable(&ch->chip->tasklet);
  }
}


/* prototypes when needed */

/* utility functions */

inline static void serocco_and_or(struct serocco_ch *ch, u8 reg, u8 and, u8 or) {
  unsigned long flags;
   local_irq_save(flags);
   serocco_writeb(ch, (serocco_readb(ch, reg)&(and))|(or), reg);
   local_irq_restore(flags);
}

static void serocco_send_cmd(struct serocco_ch *ch, u8 cmd, u8 reg,int do_schedule) {
#define TOUT 1000
  int i;
  unsigned long flags;

  local_irq_save(flags);
  for(i=0;i<TOUT;i++) {
    if (serocco_readb(ch, SEROCCO_STARL) & SEROCCO_CEC) {
      if (do_schedule)
	schedule();
      else
	udelay(1);
    }
    else
      break;
  }

  if (i==TOUT) {
    printk("BUG @ %s %d: serocco was not ready after a while, cmd is %02x to %02x\n", __FUNCTION__, __LINE__,
	   cmd, reg);
  }

  serocco_writeb(ch, cmd, reg);

  for(i=0;i<TOUT;i++) {
    if (serocco_readb(ch, reg) & cmd) {
#if 0
      /* this is broken */
      if (do_schedule)
	schedule();
      else
#endif
	udelay(1);
    }
    else
      break;
  }

  if (i==TOUT) {
    printk("BUG @ %s %d: serocco cmd didn't execute, cmd is %02x to %02x\n", __FUNCTION__, __LINE__,
	   cmd, reg);
  }

  local_irq_restore(flags);
  
#undef TOUT
}

/* pin handling functions */

#if 0
static struct serocco_pin_s * serocco_get_pin(struct serocco_ch *ch, int pin) {
  struct serocco_pin_s *r = ch->pins;

  while (r->line && r->line != pin)
    r++;
  if (!r->line) {
    printk("BUG @ %s %d: pin %d doesn't exists!\n", __FUNCTION__, __LINE__, pin);
    return NULL;
  }
  return r;
}
#endif

#if 0
static int serocco_get_pin_status(struct serocco_ch *ch, int pin) {
  struct serocco_pin_s *r = serocco_get_pin(ch, pin);

  if (!r)
    return -1;
  return ((serocco_readb(ch, r->bit_reg) & r->bit_mask) > 0) ^ r->neg;
}
#endif

static int serocco_get_all_pins_status(struct serocco_ch *ch) {
  struct serocco_pin_s *r = ch->pins;
  int ret = 0;

  while (r->line) {
    if (r->irq_reg) {
      if (((serocco_readb(ch, r->bit_reg) & r->bit_mask) > 0) ^ r->neg) {
	ret |= r->line;
      }
    }
    r++;
  }
  pr_debug("%s %d %08x\n", __FUNCTION__, __LINE__, ret);
  return ret;
}

#if 0
static int serocco_get_all_pins_events(struct serocco_ch *ch) {
  /* DO NOT USE, trashes ISR!!!! */
  struct serocco_pin_s *r = ch->pins;
  int ret = 0;

  printk("BUG @ %s %d\n", __FUNCTION__, __LINE__);
  while (r->line) {
    if (r->irq_reg) {
      if (((serocco_readb(ch, r->bit_reg) & r->bit_mask) > 0) ^ r->neg) {
	ret |= r->line;
      }
    }
    r++;
  }
  return ret;
}
#endif

#if 0
static void serocco_set_pin_status(struct serocco_ch *ch, int pin, int val) {
  struct serocco_pin_s *r = serocco_get_pin(ch, pin);

  if (!r)
    return;

  val = val ^ r->neg;
  serocco_and_or(ch, r->bit_reg, ~r->bit_mask, val ? r->bit_mask : 0);
}
#endif

static void serocco_set_all_pins_status(struct serocco_ch *ch, int s) {
  struct serocco_pin_s *r = ch->pins;

  pr_debug("%s %d %08x\n", __FUNCTION__, __LINE__, s);
  while (r->line) {
    if (!r->irq_reg) {
      serocco_and_or(ch, r->bit_reg, ~r->bit_mask, ( ( (s & r->line) > 0 ) ^ r->neg) ? r->bit_mask : 0);
    }
    r++;
  }
}

static void serocco_init_pins(struct serocco_ch *ch) {
  struct serocco_pin_s *r = ch->pins;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  while (r->line) {
    switch(r->line) {
    case TIOCM_CTS:
      serocco_and_or(ch, SEROCCO_CCR1H, ~SEROCCO_FCTS, SEROCCO_FCTS);
      break;
    case TIOCM_RTS:
      serocco_and_or(ch, SEROCCO_CCR1H, ~SEROCCO_RTS, SEROCCO_RTS);
      break;
    case TIOCM_CAR:
      serocco_and_or(ch, SEROCCO_CCR1H, ~SEROCCO_CAS, 0);
      break;
    default:
      /* we are using a GPIO */
      serocco_and_or(ch, r->bit_reg - 2,
		     ~r->bit_mask,
		     r->irq_reg ? r->bit_mask : 0);
    }
    r++;
  }
}

static void serocco_irq_on_pins(struct serocco_ch *ch, int on) {
  struct serocco_pin_s *r = ch->pins;

  pr_debug("%s %d %d\n", __FUNCTION__, __LINE__, on);
  while (r->line) {
    if (r->irq_reg) {
      serocco_and_or(ch, r->irq_reg, ~r->irq_mask , on ? 0 : r->irq_mask);
    }
    r++;
  }
}

static void serocco_calc_baud_rate(struct serocco_ch *ch, unsigned int baud,
				   u8 *brrl, u8 *brrh, unsigned long hz) {
  u8 n,m,min_n,min_m;
  int min_k, k;
  unsigned int min_delta;
  struct serocco_serial_port *up = (struct serocco_serial_port *) ch->pmode;
#define MY_ABS(x) ( ((x)>0)?(x):-(x))

  if (hz)
    min_k = hz / (16 * baud);
  else
    min_k = up->port.uartclk / (16 * baud);
  min_n = min_m = 0;
  min_delta = baud;
  for(n=0; n<64; n++) 
    for(m=0; m<16; m++) {
      unsigned int delta;

      k = (n+1)*(1<<m);
      delta = MY_ABS(k-min_k);
      if (delta < min_delta) {
	min_delta = delta;
	min_n = n;
	min_m = m;
      }
    }
  *brrl = min_n;
  *brrh = min_m;

  pr_debug("%s %d %d %d\n", __FUNCTION__, min_n, min_m, min_delta);
  
#undef MY_ABS
}


/* async linux interface */

static void serial_serocco_enable_ms(struct uart_port *port)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  serocco_irq_on_pins(up->ch,1);
}

static void serial_serocco_stop_tx(struct uart_port *port, unsigned int tty_stop)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  serocco_and_or(up->ch, SEROCCO_IMR1, 0xff, SEROCCO_XPR);
}

static void serial_serocco_stop_rx(struct uart_port *port)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  serocco_and_or(up->ch, SEROCCO_IMR0, 0xff, SEROCCO_RPF);
  up->port.read_status_mask &= ~UART_LSR_DR;
}

static void serial_serocco_tx_chars(struct serocco_serial_port * up) {
  int count;
  struct circ_buf *xmit = &up->port.info->xmit;

  if ((serocco_readb(up->ch, SEROCCO_STARL) & SEROCCO_XFW)) {
    if (up->port.x_char) {
      up->tx_empty = 0;
      serocco_writeb(up->ch, up->port.x_char, SEROCCO_XFIFOL);
      up->port.icount.tx++;
      up->port.x_char = 0;
      goto tx_end;
    }
    if (uart_circ_empty(xmit) || uart_tx_stopped(&up->port)) {
      serial_serocco_stop_tx(&up->port, 0);
      goto tx_end_no_cmd;
    }

    up->tx_empty = 0;
    count = up->port.fifosize / 2;
    do {
      serocco_writeb(up->ch, xmit->buf[xmit->tail], SEROCCO_XFIFOL);
      xmit->tail = (xmit->tail + 1) & (UART_XMIT_SIZE - 1);
      up->port.icount.tx++;
      if (uart_circ_empty(xmit))
	break;
    } while (--count > 0);

    if (uart_circ_chars_pending(xmit) < WAKEUP_CHARS) {
      uart_write_wakeup(&up->port);
    }

    if (uart_circ_empty(xmit)) {
      serial_serocco_stop_tx(&up->port, 0);
    }
    
  tx_end:
    serocco_send_cmd(up->ch, SEROCCO_XF, SEROCCO_CMDRL, 0);
  tx_end_no_cmd:
    count = count;		/* keep compiler happy */
  }  
}

static void serial_serocco_start_tx(struct uart_port *port, unsigned int tty_start)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;

  serocco_get(up->ch);
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  serocco_and_or(up->ch, SEROCCO_IMR1, ~SEROCCO_XPR, 0);  
  serial_serocco_tx_chars(up);
  serocco_put(up->ch);
  tasklet_hi_schedule(&up->ch->chip->tasklet);
}

static void serial_serocco_handle_irq(struct serocco_serial_port * up, 
				      u8 gstar, u8 gpisl, u8 gpish, 
				      u8 isr0, u8 isr1, u8 isr2) {
  struct tty_struct *tty = up->port.info->tty;
  unsigned int ch=0, flag, status=0, rstatus, cstatus;
  int count, mask;
  
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  /* flag overun, break, tx fifo empy */

  if (isr0 & SEROCCO_RFO) {
    status |= UART_LSR_OE;
  }

  if (isr1 & SEROCCO_BRK) {
    status |= UART_LSR_BI;
  }

  if (isr1 & SEROCCO_ALLS) {
    printk("ALLS report <chri@gmail.com> if you see this\n");
    up->tx_empty = 1;
  }

  /* receive chars */

  if (isr0 & (SEROCCO_RPF|SEROCCO_TIME) || status) {
    /* determine how many chars are in the buffer*/
    mask = 0x1f;
    switch(serocco_readb(up->ch, SEROCCO_CCR3H) & SEROCCO_RFTH_MASK) {
    case SEROCCO_RFTH_32: mask = 0x1f; break;
    case SEROCCO_RFTH_16: mask = 0x0f; break;
    case SEROCCO_RFTH_04: mask = 0x03; break;
    case SEROCCO_RFTH_02: mask = 0x02; break;
    }
    if (isr0 & SEROCCO_RPF) {
      count = (mask + 1) /2;
    }
    else {
      count = (serocco_readb(up->ch, SEROCCO_RBCL) & mask) / 2;
    }

    pr_debug("%s %d c %d m %d s %d\n", __FUNCTION__, __LINE__, count, mask, status);
    
    if (count == 0 && status) {
      count = 1;
    }
    
    while (count>0) {
      if (unlikely(tty->flip.count >= TTY_FLIPBUF_SIZE)) {
	if (tty->low_latency)
	  tty_flip_buffer_push(tty);
	/*
	 * If this failed then we will throw away the
	 * bytes but must do so to clear interrupts
	 */
      }
      if (isr0 & (SEROCCO_RPF|SEROCCO_TIME)) {
	status |= UART_LSR_DR;
	ch = serocco_readb(up->ch, SEROCCO_RFIFOL);
	rstatus = serocco_readb(up->ch, SEROCCO_RFIFOL);
	/* translate status from serocco to what uart layer expects */
	if (rstatus & SEROCCO_SB_PE) {
	  status |= UART_LSR_PE;
	}
	else {
	  status &= ~UART_LSR_PE;
	}
	if (rstatus & SEROCCO_SB_FE) {
	  status |= UART_LSR_FE;
	}
	else {
	  status &= ~UART_LSR_FE;
	}
      }

      if (isr1 & SEROCCO_XPR) {
	status |= UART_LSR_THRE;
      }
      if (isr1 & SEROCCO_ALLS) {
	status |= UART_LSR_TEMT;
      }

      flag = TTY_NORMAL;
      up->port.icount.rx++;

      if (unlikely(status & (UART_LSR_BI | UART_LSR_PE |
			     UART_LSR_FE | UART_LSR_OE))) {
	/*
	 * For statistics only
	 */
	if (status & UART_LSR_BI) {
	  pr_debug("%s %d BI\n", __FUNCTION__, __LINE__);
	  status &= ~(UART_LSR_FE | UART_LSR_PE);
	  up->port.icount.brk++;
	  /*
	   * We do the SysRQ and SAK checking
	   * here because otherwise the break
	   * may get masked by ignore_status_mask
	   * or read_status_mask.
	   */
	  if (uart_handle_break(&up->port))
	    goto ignore_char;
	} else if (status & UART_LSR_PE) {
	  up->port.icount.parity++;
	}
	else if (status & UART_LSR_FE) {
	  up->port.icount.frame++;
	}
	if (status & UART_LSR_OE) {
	  up->port.icount.overrun++;
	}
	/*
	 * Mask off conditions which should be ignored.
	 */
	status &= up->port.read_status_mask;

	if (status & UART_LSR_BI) {
	  flag = TTY_BREAK;
	} else if (status & UART_LSR_PE)
	  flag = TTY_PARITY;
	else if (status & UART_LSR_FE)
	  flag = TTY_FRAME;
      }

      pr_debug("%s %d %x %x %x\n", __FUNCTION__, __LINE__, status, ch, flag);
      uart_insert_char(&up->port, status, UART_LSR_OE, ch, flag);

    ignore_char:
      status = 0;
      count--;
    } 
    tty_flip_buffer_push(tty);
    serocco_send_cmd(up->ch, SEROCCO_RMC, SEROCCO_CMDRH, 0);
  }
  
  /* handle lines changes */

  status = 0;  
  /* translate to uart like values */
  rstatus = serocco_get_all_pins_status(up->ch);
  if (rstatus & TIOCM_DSR)
    status |= UART_MSR_DSR;
  if (rstatus & TIOCM_RI)
    status |= UART_MSR_RI;
  if (rstatus & TIOCM_CTS)
    status |= UART_MSR_CTS;
  if (rstatus & TIOCM_CAR)
    status |= UART_MSR_DCD;
  /* mark uart events, if any */
  pr_debug("%s %d %x %x\n", __FUNCTION__, __LINE__, rstatus, up->pin_status_in);
  cstatus = rstatus ^ up->pin_status_in;
  if (cstatus & TIOCM_DSR)
    status |= UART_MSR_DDSR;
  if (cstatus & TIOCM_RI)
    status |= UART_MSR_TERI;
  if (cstatus & TIOCM_CTS)
    status |= UART_MSR_DCTS;
  if (cstatus & TIOCM_CAR)
    status |= UART_MSR_DDCD;
  up->pin_status_in = rstatus;
  /* now standard uart code */
  if (status & UART_MSR_ANY_DELTA) {
    if (status & UART_MSR_TERI)
      up->port.icount.rng++;
    if (status & UART_MSR_DDSR)
      up->port.icount.dsr++;
    if (status & UART_MSR_DDCD)
      uart_handle_dcd_change(&up->port, status & UART_MSR_DCD);
    if (status & UART_MSR_DCTS)
      uart_handle_cts_change(&up->port, status & UART_MSR_CTS);
    
    wake_up_interruptible(&up->port.info->delta_msr_wait);
  }

  /* transmit chars */
  serial_serocco_tx_chars(up);
}

static unsigned int serial_serocco_tx_empty(struct uart_port *port)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;
  unsigned long flags;
  unsigned int ret = 0;

  pr_debug("%s %d %d\n", __FUNCTION__, __LINE__, up->tx_empty);
  spin_lock_irqsave(&up->port.lock, flags);
  if (up->tx_empty)
    ret = TIOCSER_TEMT;
  spin_unlock_irqrestore(&up->port.lock, flags);

  return ret;
}

static unsigned int serial_serocco_get_mctrl(struct uart_port *port)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;
  unsigned long flags;
  unsigned int ret;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  serocco_get(up->ch);
  spin_lock_irqsave(&up->port.lock, flags);
  ret = serocco_get_all_pins_status(up->ch);
  spin_unlock_irqrestore(&up->port.lock, flags);
  serocco_put(up->ch);
  return ret;
}

static void serial_serocco_set_mctrl(struct uart_port *port, unsigned int mctrl)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;
  
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  serocco_get(up->ch);
  serocco_set_all_pins_status(up->ch, mctrl);
  serocco_put(up->ch);
}

static void serial_serocco_break_ctl(struct uart_port *port, int break_state)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;
  unsigned long flags;

  pr_debug("%s %d %d\n", __FUNCTION__, __LINE__, break_state);
  spin_lock_irqsave(&up->port.lock, flags);
  serocco_and_or(up->ch, SEROCCO_CCR3L, ~SEROCCO_XBRK,
		 (break_state == -1) ? SEROCCO_XBRK : 0);
  spin_unlock_irqrestore(&up->port.lock, flags);
}

static int serial_serocco_startup(struct uart_port *port)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;
  u8 brrl, brrh;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  serocco_get(up->ch);
  /* initialize the port */

  /* 
      Bit clock rate 16 selected (register CCR0L, bit BCR = `1')
      Clock mode 0, 1, 3b, 4, or 7b selected (register CCR0L, bit field 'CM')
      NRZ data encoding selected (register CCR0H, bit field 'SC')

     7b clock mode
 */

  if (force_clock == -1)
    serocco_writeb(up->ch, 
		   SEROCCO_VIS | SEROCCO_BCR | SEROCCO_SSEL | SEROCCO_CM7, 
		   SEROCCO_CCR0L);
  else
    serocco_writeb(up->ch, 
		   SEROCCO_VIS | SEROCCO_BCR | force_clock, 
		   SEROCCO_CCR0L);
  serocco_writeb(up->ch, 
		 SEROCCO_POWERUP | SEROCCO_MOD_ASYNC,
		 SEROCCO_CCR0H);
  serocco_writeb(up->ch, 
		 0,
		 SEROCCO_CCR1L);
  serocco_writeb(up->ch,
		 SEROCCO_RTS | SEROCCO_FCTS,
		 SEROCCO_CCR1H);
  serocco_writeb(up->ch, 
		 FORCE_LOOPBACK,
		 SEROCCO_CCR2L);
  serocco_writeb(up->ch, 
		 0,
		 SEROCCO_CCR2H);
  serocco_writeb(up->ch, 
		 SEROCCO_RAC,
		 SEROCCO_CCR3L);
  serocco_writeb(up->ch, 
		 SEROCCO_RFDF
#ifdef BIGTO
		 |SEROCCO_RFTH_32
#endif
		 |0,
		 SEROCCO_CCR3H);

#ifdef BIGTO
  serocco_writeb(up->ch, SEROCCO_TOIE, SEROCCO_TOLEN);  
#endif
		 
  serocco_calc_baud_rate(up->ch, 115200, &brrl, &brrh, 0);
  serocco_writeb(up->ch, 
		 brrl,
		 SEROCCO_BRRL);
  serocco_writeb(up->ch, 
		 brrh,
		 SEROCCO_BRRH);
  
  serocco_init_pins(up->ch);

  serocco_send_cmd(up->ch, SEROCCO_XRES, SEROCCO_CMDRL,0);
  serocco_send_cmd(up->ch, SEROCCO_RRES, SEROCCO_CMDRH,0);
  udelay(10);

  /* enable the irqs */

  serocco_writeb(up->ch, 
		 ~(SEROCCO_RFO | 
#ifdef BIGTO
		   SEROCCO_TIME |
#endif
		   SEROCCO_RPF),
		 SEROCCO_IMR0);
  serocco_writeb(up->ch, 
		 ~(SEROCCO_CSC | SEROCCO_ALLS | SEROCCO_BRK),
		 SEROCCO_IMR1);
  serocco_writeb(up->ch, 
		 ~(SEROCCO_CDSC),
		 SEROCCO_IMR2);
  serocco_irq_on_pins(up->ch, 1);

  /* set-up misc fields of the structure */
  
  up->pin_status_in = serocco_get_all_pins_status(up->ch);
  up->tx_empty = 1;

  /* mark the port running */

  up->ch->running = 1;
  serocco_put(up->ch);
  
  return 0;
}

static void serial_serocco_shutdown(struct uart_port *port)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  up->ch->running = 0;

#ifdef DEBUG_FULL
  serocco_full_dump(up->ch);
#endif

  /* disable the irqs */

  serocco_writeb(up->ch, 
		 0xff,
		 SEROCCO_IMR0);
  serocco_writeb(up->ch, 
		 0xff,
		 SEROCCO_IMR1);
  serocco_writeb(up->ch, 
		 0x03,
		 SEROCCO_IMR2);


}

static void
serial_serocco_set_termios(struct uart_port *port, struct termios *termios,
			   struct termios *old)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;
  unsigned int baud;
  u8 brrl, brrh;
  u8 ccr3l, ccr3h;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  serocco_get(up->ch);

  ccr3l = serocco_readb(up->ch, SEROCCO_CCR3L);
  ccr3h = serocco_readb(up->ch, SEROCCO_CCR3H);

  ccr3l &= ~SEROCCO_CHL_MASK;
  switch (termios->c_cflag & CSIZE) {
  case CS5:
    ccr3l |= SEROCCO_CHL_CS5;
    break;
  case CS6:
    ccr3l |= SEROCCO_CHL_CS6;
    break;
  case CS7:
    ccr3l |= SEROCCO_CHL_CS7;
    break;
  default:
  case CS8:
    ccr3l |= SEROCCO_CHL_CS8;
    break;
  }

  if (termios->c_cflag & CSTOPB)
    ccr3l |= SEROCCO_STOP;
  else
    ccr3l &= ~SEROCCO_STOP;

  if (termios->c_cflag & PARENB)
    ccr3h |= SEROCCO_PARE;
  else
    ccr3h &= ~SEROCCO_PARE;

  ccr3h &= ~SEROCCO_PAR_MASK;
  if (!(termios->c_cflag & PARODD))
    ccr3h |= SEROCCO_PAR_ODD;
  else
    ccr3h |= SEROCCO_PAR_EVEN;

  baud = uart_get_baud_rate(port, termios, NULL, 0, port->uartclk/16);
  serocco_calc_baud_rate(up->ch, baud, &brrl, &brrh, 0);
  serocco_writeb(up->ch, 
		 brrl,
		 SEROCCO_BRRL);
  serocco_writeb(up->ch, 
		 brrh,
		 SEROCCO_BRRH);

  /*
   * Update the per-port timeout.
   */
  uart_update_timeout(port, termios->c_cflag, baud);

  up->port.read_status_mask = UART_LSR_OE | UART_LSR_THRE | UART_LSR_DR;
  if (termios->c_iflag & INPCK)
    up->port.read_status_mask |= UART_LSR_FE | UART_LSR_PE;
  if (termios->c_iflag & (BRKINT | PARMRK))
    up->port.read_status_mask |= UART_LSR_BI;

  /*
   * Characters to ignore
   */
  up->port.ignore_status_mask = 0;
  if (termios->c_iflag & IGNPAR)
    up->port.ignore_status_mask |= UART_LSR_PE | UART_LSR_FE;
  if (termios->c_iflag & IGNBRK) {
    up->port.ignore_status_mask |= UART_LSR_BI;
    /*
     * If we're ignoring parity and break indicators,
     * ignore overruns too (for real raw support).
     */
    if (termios->c_iflag & IGNPAR)
      up->port.ignore_status_mask |= UART_LSR_OE;
  }

  /*
   * ignore all characters if CREAD is not set
   */
  if ((termios->c_cflag & CREAD) == 0)
    up->port.ignore_status_mask |= UART_LSR_DR;

  /* commit changes */
  serocco_writeb(up->ch, 
		 ccr3l,
		 SEROCCO_CCR3L);
  serocco_writeb(up->ch, 
		 ccr3h,
		 SEROCCO_CCR3H);

  /*
   * CTS flow control flag and modem status interrupts
   */
  if (UART_ENABLE_MS(&up->port, termios->c_cflag))
    serocco_irq_on_pins(up->ch, 1);
  else
    serocco_irq_on_pins(up->ch, 0);

  /* reset serocco tx and rx */
  serocco_send_cmd(up->ch, SEROCCO_XRES, SEROCCO_CMDRL,0);
  serocco_send_cmd(up->ch, SEROCCO_RRES, SEROCCO_CMDRH,0);
  udelay(10);  
  serocco_put(up->ch);
}

static void
serial_serocco_pm(struct uart_port *port, unsigned int state,
		  unsigned int oldstate)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  /* TODO: serocco has suspend mode ... */
}

static void serial_serocco_release_port(struct uart_port *port)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
}

static int serial_serocco_request_port(struct uart_port *port)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  return 0;
}

static void serial_serocco_config_port(struct uart_port *port, int flags)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  up->port.type = 0x69;	/* TODO: request a proper number */
}

static int
serial_serocco_verify_port(struct uart_port *port, struct serial_struct *ser)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  /* we don't want the core code to modify any port params */
  return -EINVAL;
}

static const char *
serial_serocco_type(struct uart_port *port)
{
  struct serocco_serial_port *up = (struct serocco_serial_port *)port;
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  return up->name;
}

/* HDLC code, thanks Rodolfo! */

#define NAME "serocco_sync"
#undef PWARNING		/* undef it, just in case */
#undef PINFO		/* undef it, just in case */
#undef PERR		/* undef it, just in case */
#define PWARNING(fmt, args...) printk(KERN_WARNING NAME ": " fmt "\n" , ## args)
#define PINFO(fmt, args...) printk(KERN_INFO NAME ": " fmt "\n" , ## args)
#define PERR(fmt, args...) printk(KERN_ERR NAME ": " fmt "\n" , ## args)

static void serocco_rx_bh(struct serocco_sync_port *sp, int is_end)
{
   struct net_device *netdev = sp->dev;
   hdlc_device *hdlc = dev_to_hdlc(netdev);

   struct sk_buff *skb;
   int i, count;

   u8 tmp8;

   pr_debug("%s %d %d\n", __FUNCTION__, __LINE__, is_end);
   if (is_end) {
      /* Check the receive status byte */
      tmp8 = serocco_readb(sp->ch, SEROCCO_RSTA);
      WB(tmp8);
      if (!(tmp8&SEROCCO_VFR)) {
         PINFO("%s: rx invalid frame", netdev->name);
         hdlc->stats.rx_frame_errors++;

         goto error;
      }
      if (tmp8&SEROCCO_RDO) {
         PINFO("%s: rx data overflow", netdev->name);
         hdlc->stats.rx_over_errors++;

         goto error;
      }
      if (!(tmp8&SEROCCO_CRCOK)) {
         PINFO("%s: rx invalid CRC", netdev->name);
         hdlc->stats.rx_crc_errors++;

         goto error;
      }
      if (tmp8&SEROCCO_RAB) {
         PINFO("%s: rx message aborted by tx", netdev->name);
         hdlc->stats.rx_frame_errors++;

         goto error;
      }

      count = (serocco_readb(sp->ch, SEROCCO_RBCH)<<8|
               serocco_readb(sp->ch, SEROCCO_RBCL))-sp->rx_index;
   }
   else
      count = 32;   /* FIXME: can be variable */

   if (sp->rx_index == 0) {
      sp->rx_skbuff = dev_alloc_skb(SEROCCO_MTU);
      RXS;
      if (!sp->rx_skbuff) {
         PERR("%s: unable to allocate packet room!", netdev->name);
         hdlc->stats.rx_dropped++;
         return;
      }
      sp->rx_skbuff->dev = netdev;
      skb_reserve(sp->rx_skbuff, 2);   /* 16 byte allign the IP fields */
      sp->rx_skbuff->protocol = htons(ETH_P_IP);
   }
   skb = sp->rx_skbuff;

   /* Make room (update also skb->len) */
   if (count > 32)   /* FIXME: can be variable */
      PWARNING("SEROCCO-H packet packet too long! Possible data corrupt");
   skb_put(skb, count);

   /* Read data from RFIFO */
   for (i = 0; i < count; i++)
      skb->data[sp->rx_index+i] = serocco_readb(sp->ch, SEROCCO_RFIFOL);
   sp->rx_index += count; 

   hdlc->stats.rx_bytes += count;

   if (skb->len > 2000){
     printk("%s: received too long packet\n", netdev->name);
     goto error;
   }

   if (is_end) {
      pr_debug("%s: packet received (%d bytes, status %02x)\n", netdev->name, sp->rx_index, skb->data[skb->len - 1]);
#if !defined(DEBUG) && defined(SHOWP)  
      printk("%s: packet received (%d bytes, status %02x)\n", netdev->name, sp->rx_index, skb->data[skb->len - 1]);
#endif

      netdev->last_rx = jiffies;

      skb->ip_summed = CHECKSUM_UNNECESSARY;
      skb->protocol = hdlc_type_trans(skb, netdev);

      skb_pull(skb, 2);
      skb_trim(skb, skb->len - 1); /* this is status char (see RSTA) */

      DUMPP(skb);

      /* Send the packet to the upper layer without removing the HDLC address */
      RXI(skb->len);
      netif_rx(skb);

      hdlc->stats.rx_packets++;

      sp->rx_index = 0;
   }

   /* Notify receive message complete */
   serocco_send_cmd(sp->ch, SEROCCO_RMC, SEROCCO_CMDRH, 0);

   return;

error:
   pr_debug("%s %d\n", __FUNCTION__, __LINE__);
   RXI(sp->rx_skbuff->len);
   //DUMPP(sp->rx_skbuff);
   if (sp->rx_index > 0)
     kfree_skb(sp->rx_skbuff);
   sp->rx_index = 0;

   hdlc->stats.rx_errors++;

   serocco_send_cmd(sp->ch, SEROCCO_RMC, SEROCCO_CMDRH, 0);
}

static void serocco_tx_bh(struct serocco_sync_port *sp)
{
   struct net_device *netdev = sp->dev;
   hdlc_device *hdlc = dev_to_hdlc(netdev);

   int i, count;

   if (!sp->tx_skbuff) {   /* nothing to tx */
      pr_debug("%s: uh! Nothing to tx??\n", netdev->name);

      netif_start_queue(netdev);
      STARTI;

      return;
   }
   pr_debug("%s %d %d %d\n", __FUNCTION__, __LINE__, sp->tx_index, sp->tx_skbuff->len);

   /* Check if we have finished to transmit the packet */
   if (sp->tx_skbuff->len == sp->tx_index) {
      kfree_skb(sp->tx_skbuff);
      sp->tx_skbuff = NULL;

      TXI;

      return;
   }

   /* Sanity check... */
   if (!(serocco_readb(sp->ch, SEROCCO_STARL)&SEROCCO_XFW)) {
      PERR("%s: uh! Why we are here if XFIFO are _not_ ready??", netdev->name);

      netif_start_queue(netdev);
      STARTI;

      return;
   }

   count = sp->tx_skbuff->len - sp->tx_index;
   if (count > 32)
      count = 32;   /* FIXME: may be variable */
      
   /* Write data into XFIFO */
   for (i = 0; i < count; i++)
      serocco_writeb(sp->ch, sp->tx_skbuff->data[sp->tx_index + i], SEROCCO_XFIFOL);

   sp->tx_index += count;
   hdlc->stats.tx_bytes += count;

   /* We are sending the last serocco packet? */
   if (sp->tx_skbuff->len == sp->tx_index)
      serocco_send_cmd(sp->ch, SEROCCO_XF|SEROCCO_XME, SEROCCO_CMDRL, 0);
   else
      serocco_send_cmd(sp->ch, SEROCCO_XF, SEROCCO_CMDRL, 0);
}

static void serocco_sync_interrupt_channel(struct serocco_sync_port *sp, u8 isr0, u8 isr1, u8 isr2) {
  struct net_device *netdev = sp->dev;
  hdlc_device *hdlc = dev_to_hdlc(netdev);
  u8 tmp;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
      
  if (isr0&SEROCCO_RME) {
    serocco_rx_bh(sp, 1);
  }
  if (isr0&SEROCCO_RPF) {
    serocco_rx_bh(sp, 0);
  }
  if (isr1&SEROCCO_XPR) {
    serocco_tx_bh(sp);
  }
  if (isr1&SEROCCO_ALLS) {
    hdlc->stats.tx_packets++;

    netif_wake_queue(netdev);
    STARTI;
    TXA;
  }
  /* Carrier detect status */
  if (isr2&SEROCCO_CDSC) {
    tmp = serocco_readb(sp->ch, SEROCCO_STARH)&SEROCCO_CD;
    if (tmp != sp->last_carrier_status) {
      sp->last_carrier_status = tmp;
      PINFO("%s: carrier status is %s", netdev->name, tmp ? "low" : "high");
    }
  } 

  /* --- Error interrupts -------------------------------------------------- */
  if (isr0&SEROCCO_FLEX) {
    PWARNING("%s: rx frame too long", netdev->name);
    hdlc->stats.rx_length_errors++;
    hdlc->stats.rx_errors++;
    
    sp->rx_index = 0;
    kfree_skb(sp->rx_skbuff);
    sp->rx_skbuff = NULL;
  }

  if (isr1&SEROCCO_XDU) {
    PWARNING("%s: tx underrun", netdev->name);
    hdlc->stats.tx_errors++;

    /* Reset the tx index, reset and (re)start transmission */
    sp->tx_index = 0;
    //serocco_tx_bh(sp);
    serocco_send_cmd(sp->ch, SEROCCO_XRES, SEROCCO_CMDRL,0);
  }
}

static void serocco_reset_txrx_data(struct serocco_sync_port *sp)
{
   struct net_device *netdev = sp->dev;
   hdlc_device *hdlc = dev_to_hdlc(netdev);

   WW(__FUNCTION__);
   pr_debug("%s %d\n", __FUNCTION__, __LINE__);

   /* Clean up tx/rx indexes */
   sp->tx_index = 0;
   sp->rx_index = 0;

   if (sp->tx_skbuff) {
      hdlc->stats.tx_dropped++; 

      kfree_skb(sp->tx_skbuff);
      sp->tx_skbuff = NULL;
   }
   if (sp->rx_skbuff && sp->rx_index > 0) {
      hdlc->stats.rx_dropped++; 

      kfree_skb(sp->rx_skbuff);
      sp->rx_skbuff = NULL;
   }
}

static void serocco_enable_irqs(struct serocco_sync_port *sp)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);

  /* We don't care about GPP interrupts CHRI .... check this!*/
  serocco_writeb(sp->ch, 0x07, SEROCCO_GPIML);
  serocco_writeb(sp->ch, 0xff, SEROCCO_GPIMH);


  /* Enable all interrupts */
  serocco_writeb(sp->ch, 0x00, SEROCCO_IMR0);
  serocco_writeb(sp->ch, 0x01, SEROCCO_IMR1);
  serocco_writeb(sp->ch, 0x00, SEROCCO_IMR2);
}

static void serocco_disable_irqs(struct serocco_sync_port *sp)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);

  /* We don't care about GPP interrupts */
  serocco_writeb(sp->ch, 0x07, SEROCCO_GPIML);
  serocco_writeb(sp->ch, 0xff, SEROCCO_GPIMH);

  /* Enable all interrupts */
  serocco_writeb(sp->ch, 0xff, SEROCCO_IMR0);
  serocco_writeb(sp->ch, 0xff, SEROCCO_IMR1);
  serocco_writeb(sp->ch, 0x03, SEROCCO_IMR2);
}

static void serocco_setup(struct serocco_sync_port *sp) {
	u8 brrl, brrh;

	/* clock mode */
	switch (sp->clock_mode) {
	case CLOCK_EXT:
		serocco_writeb(sp->ch, SEROCCO_VIS, SEROCCO_CCR0L);           /* Set mode 0a */
		break;
	case CLOCK_TXINT:
	case CLOCK_TXFROMRX:	
		printk("(Not correctly implemented fall-back to internal) ");
	case CLOCK_INT:	
		serocco_writeb(sp->ch, 
			       SEROCCO_VIS | SEROCCO_TOE  | SEROCCO_SSEL | SEROCCO_CM7, 
			       SEROCCO_CCR0L);           /* Set mode 7b */		
		/* clock rate */
		serocco_calc_baud_rate(sp->ch, sp->clock_rate / 16, &brrl, &brrh, UART_CLOCK);
		serocco_writeb(sp->ch, 
			       brrl,
			       SEROCCO_BRRL);
		serocco_writeb(sp->ch, 
			       brrh,
			       SEROCCO_BRRH);  
	}
	/* encoding */
	switch (sp->encoding) {
	case ENCODING_NRZ :
		serocco_writeb(sp->ch, SEROCCO_POWERUP | SEROCCO_BUS_NRZ, SEROCCO_CCR0H);
		break;
	case ENCODING_NRZI : 
		serocco_writeb(sp->ch, SEROCCO_POWERUP | SEROCCO_BUS_NRZI, SEROCCO_CCR0H);
		break;
	case ENCODING_FM_MARK : 
		serocco_writeb(sp->ch, SEROCCO_POWERUP | SEROCCO_BUS_FM0, SEROCCO_CCR0H);
		break;
	case ENCODING_FM_SPACE : 
		serocco_writeb(sp->ch, SEROCCO_POWERUP | SEROCCO_BUS_FM1, SEROCCO_CCR0H);
		break;
	case ENCODING_MANCHESTER : 
		serocco_writeb(sp->ch, SEROCCO_POWERUP | SEROCCO_BUS_MANCHESTER, SEROCCO_CCR0H);
		break;
	}
	/* parity */
	switch (sp->parity) {
	case PARITY_NONE:
	case PARITY_CRC16_PR0:
	case PARITY_CRC16_PR1:
		printk("(Not correctly handled, falling back to PARITY_CRC16_PR1_CCITT) ");
	case PARITY_CRC16_PR1_CCITT :
		serocco_writeb(sp->ch, (u8) ~(SEROCCO_CRL|SEROCCO_C32), SEROCCO_CCR1L);
		break;
	case PARITY_CRC16_PR0_CCITT:
		serocco_writeb(sp->ch, (u8) ~(SEROCCO_C32), SEROCCO_CCR1L);
		break;
	case PARITY_CRC32_PR0_CCITT:
		serocco_writeb(sp->ch, (u8) ~(0), SEROCCO_CCR1L);
		break;
	case PARITY_CRC32_PR1_CCITT:
		serocco_writeb(sp->ch, (u8) ~(SEROCCO_CRL), SEROCCO_CCR1L);
		break;
	}
	/* loopback */
	switch(sp->loopback) {
	case SSMODE_LOOP_ON:
		serocco_writeb(sp->ch, SEROCCO_MDS_ADDR_01|
#if SEROCCO_ADDRESS_MODE == 1
			       SEROCCO_ADM|
#endif
			       3, SEROCCO_CCR2L);   /* Setting Automode 0 or 1 */
		
		break;
	case SSMODE_LOOP_OFF:
		serocco_writeb(sp->ch, SEROCCO_MDS_ADDR_01|
#if SEROCCO_ADDRESS_MODE == 1
			       SEROCCO_ADM|
#endif
			       0, SEROCCO_CCR2L);   /* Setting Automode 0 or 1 */
		
		break;
	}

	serocco_and_or(sp->ch, SEROCCO_CCR1L, SEROCCO_CRL|SEROCCO_C32, SEROCCO_ODS); /* TxD push/pull */

#ifdef DEBUG
	PRREG(sp->ch, SEROCCO_CCR0L);
	PRREG(sp->ch, SEROCCO_CCR0H);
	PRREG(sp->ch, SEROCCO_CCR1L);
	PRREG(sp->ch, SEROCCO_CCR1H);
	PRREG(sp->ch, SEROCCO_CCR2L);
	PRREG(sp->ch, SEROCCO_CCR2H);
	PRREG(sp->ch, SEROCCO_CCR3L);
	PRREG(sp->ch, SEROCCO_CCR3H);
#endif

	serocco_send_cmd(sp->ch, SEROCCO_RRES, SEROCCO_CMDRH,0);        /* reset receiver to set new parameters*/
}

static void serocco_up(struct serocco_sync_port *sp)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);

  WW(__FUNCTION__);
  serocco_reset_txrx_data(sp);

  serocco_enable_irqs(sp);

  /* We do not reset the transmitter here to avoid getting an
     unneeded interrupt */

  serocco_setup(sp);

  serocco_send_cmd(sp->ch, SEROCCO_RRES, SEROCCO_CMDRH,0);        /* reset receiver */
  serocco_and_or(sp->ch, SEROCCO_CCR3L, ~0, SEROCCO_RAC);     /* enable rx */

  //serocco_send_cmd(sp->ch, SEROCCO_XRES, SEROCCO_CMDRL,0);
}

static void serocco_down(struct serocco_sync_port *sp)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);

  WW(__FUNCTION__);
  serocco_send_cmd(sp->ch, SEROCCO_XRES, SEROCCO_CMDRL,0);      /* abort current tx */
  serocco_and_or(sp->ch, SEROCCO_CCR3L, ~SEROCCO_RAC, 0);   /* abort further rx */

  serocco_reset_txrx_data(sp);

  serocco_disable_irqs(sp);

  serocco_and_or(sp->ch, SEROCCO_CCR0H, (u8) ~SEROCCO_POWERUP, 0);  /* power down */
}

/* ----- Set iface operations ----------------------------------------- */

static int serocco_encoding_setting(struct net_device *netdev)
{
  pr_debug("%s: serocco_encoding_setting\n", netdev->name);  
  return 0;
}

static int serocco_clock_setting(struct net_device *netdev)
{
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;
  pr_debug("%s: serocco_clock_setting\n", netdev->name);

  sp->clock_rate = sp->settings.clock_rate;
  sp->clock_mode = sp->settings.clock_type;

  return 0;
}

static int serocco_loopback_setting(struct net_device *netdev)
{
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;
  pr_debug("%s: serocco_loopback_setting\n", netdev->name);

  sp->loopback = sp->settings.loopback;

  return 0;
}

static int serocco_crc_setting(struct net_device *netdev)
{
  //struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;
  pr_debug("%s: serocco_crc_setting\n", netdev->name);
  return 0;
}

static int serocco_set_iface(struct net_device *netdev)
{
  struct {
    int (*action)(struct net_device *);
  } *p, do_setting[] = {
    { serocco_encoding_setting }, { serocco_clock_setting },
    { serocco_loopback_setting }, { serocco_crc_setting },
    { NULL }
  };
  int ret = 0;
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;

  pr_debug("%s: serocco_set_iface\n", netdev->name);

  for (p = do_setting; p->action; p++) {
    if ((ret = p->action(netdev)) < 0)
      break;
  }

  serocco_setup(sp);

  return ret;
}

/* ----- File operations ---------------------------------------------- */

void serocco_tx_timeout(struct net_device *netdev)
{
  hdlc_device *hdlc = dev_to_hdlc(netdev);
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;

  WW(__FUNCTION__);
  serocco_down(sp);

  PWARNING("%s: transmit timeout at %ld, latency %ld",
	   netdev->name, jiffies, jiffies-netdev->trans_start);

  hdlc->stats.tx_errors++;

  serocco_up(sp);
  netif_start_queue(netdev);
  STARTI;

  return;
}

static int serocco_hdlc_xmit(struct sk_buff *skb, struct net_device *netdev)
{
  hdlc_device *hdlc = dev_to_hdlc(netdev);
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);

  TXT(skb->len);
    
  /* If we are here and tx_skbuff is not free then tx queue is full */
  if (sp->tx_skbuff) {
    PERR("%s: tx queue full!", netdev->name);

    /* No space in tx ring so we free socket buffer and we have to stop
       the device queue */
    netif_stop_queue(netdev);
    STOPI;

    hdlc->stats.tx_dropped++;

    return 1;
  }


  /* Enqueue new packet */
  sp->tx_skbuff = skb_copy(skb, GFP_ATOMIC);
  if (!sp->tx_skbuff) {
    /* shouldn't come here! */
    PERR("%s: unable to copy HDLC packet!", netdev->name);
    netif_stop_queue(netdev);
    STOPI;

    hdlc->stats.tx_dropped++;

    return 1;
  }
  sp->tx_index = 0;

  DUMPP(skb);
  TXS(skb->len);

  /* Add HDLC address */
  skb_push(sp->tx_skbuff, 2);

  /* Init tx data */
  netdev->trans_start = jiffies;
  sp->tx_skbuff->dev = netdev;

  /* Signal SEROCCO-H to start transmission */
  //serocco_tx_bh(sp);
  serocco_send_cmd(sp->ch, SEROCCO_XRES, SEROCCO_CMDRL,0);  

  netif_stop_queue(netdev);
  STOPI;

  return 0;
}

static int serocco_ioctl(struct net_device *netdev, struct ifreq *ifr, int cmd)
{
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;
  struct if_settings *if_s = &ifr->ifr_settings;
  const size_t size = sizeof(sp->settings);
  int ret = 0;

  pr_debug("%s: serocco_ioctl 0x%08x\n", netdev->name, ifr->ifr_settings.type);

  switch(ifr->ifr_settings.type) {
#if 1
  case IF_GET_IFACE : {
    if_s->type = IF_IFACE_SYNC_SERIAL;

    if (if_s->size == 0)
      return 0;
    if (if_s->size < size)
      return -ENOMEM;
    if (copy_to_user(if_s->ifs_ifsu.sync, &sp->settings, size))
      return -EFAULT;

    if_s->size = size;

    break;
  }
    
  case IF_IFACE_SYNC_SERIAL : {
    if (!capable(CAP_NET_ADMIN))
      return -EPERM;
    if (if_s->size != size)
      return -ENOMEM;
    if (copy_from_user(&sp->settings, if_s->ifs_ifsu.sync, size))
      return -EFAULT;

    ret = serocco_set_iface(netdev);

    break;
  }
#endif

  default: {
    ret = hdlc_ioctl(netdev, ifr, cmd);
  }
  }

  return ret;
}

static int serocco_hdlc_attach(struct net_device *netdev, unsigned short encoding, unsigned short parity)
{
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;

  pr_debug("%s: serocco_hdlc_attach\n", netdev->name);

  sp->encoding = encoding;
  sp->parity = parity;
  
#if FORCE_LOOPBACK == 0
  sp->loopback = SSMODE_LOOP_OFF;
#else
  sp->loopback = SSMODE_LOOP_ON;
#endif


  if (sp->encoding == ENCODING_DEFAULT)
	  encoding = ENCODING_NRZ;

  if (sp->encoding == PARITY_DEFAULT)
	  parity = PARITY_CRC16_PR1_CCITT;

  serocco_and_or(sp->ch, SEROCCO_CCR0H, ~0, SEROCCO_POWERUP); /* power up */

  /* Init the SEROCCO-H according to supplied configuration */
  /* See data sheet at page 5-149 */
  serocco_writeb(sp->ch, SEROCCO_MDS_ADDR_01|
#if SEROCCO_ADDRESS_MODE == 1
		 SEROCCO_ADM|
#endif
		 FORCE_LOOPBACK|
		 0, SEROCCO_CCR2L);   /* Setting Automode 0 or 1 */
  serocco_writeb(sp->ch, 0, SEROCCO_CCR3L);
  serocco_writeb(sp->ch, SEROCCO_RADD, SEROCCO_CCR3H); /* HDLC addr is forwarded */
#if SEROCCO_ADDRESS_MODE == 1
  serocco_writeb(sp->ch, (address>>8)&0xff, SEROCCO_RAH1);
  pr_debug("HDLC address 1 = %x\n", (address>>8)&0xff);
  serocco_writeb(sp->ch, address&0x00ff, SEROCCO_RAH2);
  pr_debug("HDLC address 2 = %x\n", address&0x00ff);
#endif

  switch (sp->encoding) {
  case ENCODING_NRZ : {
    pr_debug("%s: encoding: NRZ\n", netdev->name);
    //serocco_writeb(sp->ch, SEROCCO_POWERUP | SEROCCO_BUS_NRZ_TIME2, SEROCCO_CCR0H);
    serocco_writeb(sp->ch, SEROCCO_POWERUP | SEROCCO_BUS_NRZ, SEROCCO_CCR0H);
    break;
  }
  }
  switch (sp->parity) {
  case PARITY_CRC16_PR1_CCITT : {
    pr_debug("%s: parity: CRC16 PR1 CCITT\n", netdev->name);
    serocco_writeb(sp->ch, (u8) ~(SEROCCO_CRL|SEROCCO_C32), SEROCCO_CCR1L);
    break;
  }
  }
  serocco_and_or(sp->ch, SEROCCO_CCR1L, SEROCCO_CRL|SEROCCO_C32, SEROCCO_ODS); /* TxD push/pull */

  //serocco_and_or(sp->ch, SEROCCO_CCR1L, 0xff, SEROCCO_DIV);

  if (force_master == 1) {
    u8 brrl, brrh;

    printk("default clock to mode 7b, baud %d\n", force_baud);
    serocco_writeb(sp->ch, 
		   SEROCCO_VIS | SEROCCO_TOE  | SEROCCO_SSEL | SEROCCO_CM7, 
		   SEROCCO_CCR0L);           /* Set mode 7b */

    serocco_calc_baud_rate(sp->ch, force_baud / 16, &brrl, &brrh, UART_CLOCK);
    serocco_writeb(sp->ch, 
		   brrl,
		   SEROCCO_BRRL);
    serocco_writeb(sp->ch, 
		   brrh,
		   SEROCCO_BRRH);  
    sp->clock_mode = CLOCK_INT;
    sp->clock_rate = force_baud;
  }
  else {
    serocco_writeb(sp->ch, SEROCCO_VIS, SEROCCO_CCR0L);           /* Set mode 0a */
    sp->clock_mode = CLOCK_EXT;
  }
  printk("WARNING, forcing ignoring CTS,RTS,CD\n");
  serocco_writeb(sp->ch,
		 SEROCCO_ICD | SEROCCO_RTS | SEROCCO_FCTS,
		 SEROCCO_CCR1H);
  serocco_writeb(sp->ch,
		 0,
		 SEROCCO_CCR2H);
  

  pr_debug("%s: %d\n", __FUNCTION__, __LINE__);

  /* Additional init steps */
#if SEROCCO_MTU >= 2048
#  error you cannot use a SEROCCO_MTU equal or greater then 2048
#endif

  serocco_writeb(sp->ch, (SEROCCO_MTU-16)&0x0ff, SEROCCO_RLCRL); /* en rx check len */
  serocco_writeb(sp->ch, (((SEROCCO_MTU-16)>>8)&0x007)|SEROCCO_RCE, SEROCCO_RLCRH);

  pr_debug("%s: done\n", __FUNCTION__);
  return 0;
}

static int serocco_open(struct net_device *netdev)
{
  hdlc_device *hdlc = dev_to_hdlc(netdev);
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;
  int ret;

  pr_debug("%s: serocco_open\n", netdev->name);

  if (sp->running) {
    PWARNING("%s: already running", netdev->name);
    return 0;
  }
  sp->running = 1;
  sp->ch->running = 1;

  /* Check the open method */
  if (hdlc->proto.open) {
    ret = hdlc->proto.open(netdev);
    if (ret < 0)
      return ret;
  } 

  serocco_up(sp);

  netif_start_queue(netdev);
  STARTI;

  return 0;
}

static int serocco_close(struct net_device *netdev)
{
  hdlc_device *hdlc = dev_to_hdlc(netdev);
  struct serocco_sync_port *sp = (struct serocco_sync_port *) dev_to_hdlc(netdev)->priv;

  pr_debug("%s: serocco_close\n", netdev->name);

  sp->running = 0;
  sp->ch->running = 1;

  netif_stop_queue(netdev);
  STOPI;

  /* Check the HDLC stop method */
  if (hdlc->proto.stop)
    hdlc->proto.stop(netdev);         

  serocco_down(sp);

  return 0;
}

static int serocco_sync_init(struct net_device *netdev)
{
   hdlc_device *hdlc = dev_to_hdlc(netdev);

   pr_debug("%s: init\n", netdev->name);

   /* Init the device structs */
   netdev->open = serocco_open;
   netdev->stop = serocco_close;
   netdev->do_ioctl = serocco_ioctl;
   netdev->tx_timeout = serocco_tx_timeout;
   netdev->watchdog_timeo = 5*HZ;   /* FIXME: should be configurable */
   netdev->tx_queue_len = 100;
   netdev->flags = IFF_NOARP;
   hdlc->xmit = serocco_hdlc_xmit;
   hdlc->attach = serocco_hdlc_attach;
   //hdlc->proto = IF_PROTO_HDLC;

   return 0;
}


#undef PWARNING		/* undef it, just in case */
#undef PINFO		/* undef it, just in case */
#undef PERR		/* undef it, just in case */
#undef NAME

/* generic irq handling */

static void serocco_handle_ch(struct serocco_ch *ch, u8 gstar, u8 gpisl, u8 gpish) {
  u8 isr0, isr1, isr2;

  isr0 = serocco_readb(ch, SEROCCO_ISR0);
  isr1 = serocco_readb(ch, SEROCCO_ISR1);
  isr2 = serocco_readb(ch, SEROCCO_ISR2);

  pr_debug("%s %d %02x %02x %02x\n", __FUNCTION__, __LINE__, isr0, isr1, isr2);

  if (ch->running) {
    switch (ch->mode & SEROCCO_MODE_MASK) {
    case SEROCCO_MODE_ASYNC: {
      struct serocco_serial_port *up = (struct serocco_serial_port *) ch->pmode;
      
      serial_serocco_handle_irq(up, gstar, gpisl, gpish, isr0, isr1, isr2);
    }
      break;
    case SEROCCO_MODE_SYNC: {
      struct serocco_sync_port *sp = (struct serocco_sync_port *) ch->pmode;

      serocco_sync_interrupt_channel(sp, isr0, isr1, isr2);
    }
      break;
    }
  }
}

static void serocco_tasklet(unsigned long chip_) {
  struct serocco_chip *chip = (struct serocco_chip *) chip_;

  u8 gstar, gpisl,gpish;

  gstar = serocco_readb(chip->ch_a, SEROCCO_GSTAR);
  while (gstar) {
    gpisl = serocco_readb(chip->ch_a, SEROCCO_GPISL);
    gpish = serocco_readb(chip->ch_a, SEROCCO_GPISH);

    pr_debug("%s %d %08x %08x %08x\n", __FUNCTION__, __LINE__, gstar, gpisl, gpish);

    serocco_handle_ch(chip->ch_a, gstar, gpisl, gpish);
    serocco_handle_ch(chip->ch_b, gstar, gpisl, gpish);

    gstar = serocco_readb(chip->ch_a, SEROCCO_GSTAR);
  }
}

static irqreturn_t
serocco_irq(int irq, void *dev_id, struct pt_regs *regs)
{
  struct serocco_chip *c = (struct serocco_chip *) dev_id;

  pr_debug("!\n");
  tasklet_hi_schedule(&c->tasklet);
  return IRQ_HANDLED;
}


/* registration of linux async interface */

struct uart_ops serial_serocco_pops = {
	.tx_empty	= serial_serocco_tx_empty,
	.set_mctrl	= serial_serocco_set_mctrl,
	.get_mctrl	= serial_serocco_get_mctrl,
	.stop_tx	= serial_serocco_stop_tx,
	.start_tx	= serial_serocco_start_tx,
	.stop_rx	= serial_serocco_stop_rx,
	.enable_ms	= serial_serocco_enable_ms,
	.break_ctl	= serial_serocco_break_ctl,
	.startup	= serial_serocco_startup,
	.shutdown	= serial_serocco_shutdown,
	.set_termios	= serial_serocco_set_termios,
	.pm		= serial_serocco_pm,
	.type		= serial_serocco_type,
	.release_port	= serial_serocco_release_port,
	.request_port	= serial_serocco_request_port,
	.config_port	= serial_serocco_config_port,
	.verify_port	= serial_serocco_verify_port,
};

static struct serocco_serial_port serial_serocco_ports[] =
  {
    {
      .name	= "SEROCCO0",
      .port	= {
	.type		= 0x69,	/* TODO: assign number */
	.iotype		= UPIO_MEM,
	.membase	= (void *) 0,
	.mapbase	= 0,
	.irq		= 0,
	.uartclk	= UART_CLOCK,
	.fifosize	= 64,
	.ops		= &serial_serocco_pops,
	.line		= 0,
      },
    },
    {
      .name	= "SEROCCO1",
      .port	= {
	.type		= 0x69,	/* TODO: assign number */
	.iotype		= UPIO_MEM,
	.membase	= (void *) 0,
	.mapbase	= 0,
	.irq		= 0,
	.uartclk	= UART_CLOCK,
	.fifosize	= 64,
	.ops		= &serial_serocco_pops,
	.line		= 1,
      },
    },
  };

static struct uart_driver serial_serocco_reg = 
{
	.owner		= THIS_MODULE,
	.driver_name	= "Serocco serial",
	.devfs_name	= "tts/",
	.dev_name	= "ttySER",
	.major		= TTY_MAJOR,
	.minor		= 128,
	.nr		= 1,
};

static int serocco_suspend(struct device *_dev, pm_message_t state, u32 level)
{
  int cur_port = 0;
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  
  for(cur_port = 0; cur_port < 2; cur_port++) {
	  struct serocco_ch *ch;
	  
	  if (!cur_port)
		  ch = my_serocco->ch_a;
	  else
		  ch = my_serocco->ch_b;
		  
	  switch(ch->mode) {
	  case SEROCCO_MODE_ASYNC:
	  {
		  struct serocco_serial_port *sport = (struct serocco_serial_port *) ch->pmode;

		  if (sport && level == SUSPEND_DISABLE)
			  uart_suspend_port(&serial_serocco_reg, &sport->port);
	  }
	  break;
	  }
  }

  return 0;
}

static int serocco_resume(struct device *_dev, u32 level)
{
  int cur_port = 0;
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);

  for(cur_port = 0; cur_port < 2; cur_port++) {
	  struct serocco_ch *ch;
	  
	  if (!cur_port)
		  ch = my_serocco->ch_a;
	  else
		  ch = my_serocco->ch_b;
		  
	  switch(ch->mode) {
	  case SEROCCO_MODE_ASYNC:
	  {
		  struct serocco_serial_port *sport = (struct serocco_serial_port *) ch->pmode;

		  if (sport && level == RESUME_ENABLE)
			  uart_resume_port(&serial_serocco_reg, &sport->port);
	  }
	  }
  }
  return 0;
}

/* generic routines for probing */

static int serocco_probe(struct device *_dev)
{
  u32 data = 0;
  int cur_asyn_port = 0, cur_port = 0;

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  
  /* initialize serocco strucure */

#define ALLOC_ONE(P,S) \
	P = kmalloc(sizeof(S), GFP_KERNEL); \
	if (!P) { \
	  printk("kalloc failed at %s %d\n", __FUNCTION__, __LINE__); \
	  goto mem_alloc; \
	} \
	memset(P,0,sizeof(S))
	

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  ALLOC_ONE(my_serocco, struct serocco_chip);
  init_MUTEX(&my_serocco->lock);
  my_serocco->base_p = SEROCCO_PHYS;
  my_serocco->base = (unsigned long) ioremap_nocache(SEROCCO_PHYS, SEROCCO_SIZE);
  if (!my_serocco->base) {
    printk("ioremap failed!\n");
    goto mem_alloc;
  }

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  ALLOC_ONE(my_serocco->ch_a, struct serocco_ch);
  ALLOC_ONE(my_serocco->ch_b, struct serocco_ch);
  my_serocco->ch_a->chip = my_serocco->ch_b->chip = my_serocco;

  my_serocco->ch_a->off = 0;
  my_serocco->ch_a->mode = serocco_port_a_mode;
  my_serocco->ch_a->pins = serocco_pins_port_a;

  my_serocco->ch_b->off = 0x50;
  my_serocco->ch_b->mode = serocco_port_b_mode;
  my_serocco->ch_b->pins = serocco_pins_port_b;

#undef ALLOC_ONE

  /* reset serocco */

#ifdef SEROCCO_RESET_PIN
  pr_debug("%s %d reset\n", __FUNCTION__, __LINE__);
  pxa_gpio_mode(SEROCCO_RESET_PIN | GPIO_OUT);
  GPDR2 |= GPIO_bit(SEROCCO_RESET_PIN);
  GPCR2 |= GPIO_bit(SEROCCO_RESET_PIN);  /* init 0 */
  mdelay(10);
  GPSR2 |= GPIO_bit(SEROCCO_RESET_PIN);  /* init 1 */	
  mdelay(10);
#endif

  /* sw reset */

  serocco_writeb(my_serocco->ch_a, SEROCCO_SWR, SEROCCO_GCMDR);
  (void) serocco_readb(my_serocco->ch_a, SEROCCO_GCMDR);
  mdelay(10);
  serocco_writeb(my_serocco->ch_a, 0, SEROCCO_GCMDR);
  mdelay(10);  

  /* check the presence of serocco */

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  data |= serocco_readb(my_serocco->ch_a, SEROCCO_VER0);
  data |= serocco_readb(my_serocco->ch_a, SEROCCO_VER1)<<8;
  data |= serocco_readb(my_serocco->ch_a, SEROCCO_VER2)<<16;
  data |= serocco_readb(my_serocco->ch_a, SEROCCO_VER3)<<24;

  if (data != 0x3005e083) {
    printk("Unsupported Serocco found, ver is %x\n", data);
    //goto no_serocco;
  }

  /* register IRQ and tasklet */

  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  tasklet_init(&my_serocco->tasklet, serocco_tasklet, (unsigned long) my_serocco);
	
  serocco_writeb(my_serocco->ch_a, SEROCCO_OSCPD | SEROCCO_DSHP | SEROCCO_IPC_LOW, SEROCCO_GMODE);
  my_serocco->irq = IRQ_GPIO(SEROCCO_IRQ_PIN);
	
  if (request_irq(my_serocco->irq, serocco_irq,
		  0, "serocco", my_serocco)) {
    printk("request_irq SEROCCO ERROR\n");
    goto no_serocco;
  }
  set_irq_type(my_serocco->irq, IRQT_FALLING);	
	

  /* ok, serocco present, register ports depending on type */
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  
  if (my_serocco->ch_a->mode == SEROCCO_MODE_ASYNC ||
      my_serocco->ch_b->mode == SEROCCO_MODE_ASYNC) {
	  int ret;
      
	  if (my_serocco->ch_a->mode == SEROCCO_MODE_ASYNC &&
	      my_serocco->ch_b->mode == SEROCCO_MODE_ASYNC)
		  serial_serocco_reg.nr = 2;

	  pr_debug("%s %d n serials %d: %d %d\n", __FUNCTION__, __LINE__, serial_serocco_reg.nr,
		   serial_serocco_ports[0].port.line,
		   serial_serocco_ports[1].port.line
		   );
	  
	  ret = uart_register_driver(&serial_serocco_reg);
	  if (ret != 0) {
		  printk("Cannot register serial driver");
		  goto no_serocco;
	  }
  }

  for(cur_port = 0; cur_port < 2; cur_port++) {
	  struct serocco_ch *ch;
	  
	  if (!cur_port)
		  ch = my_serocco->ch_a;
	  else
		  ch = my_serocco->ch_b;
		  
	  switch(ch->mode) {
	  case SEROCCO_MODE_ASYNC:
	  {
		  int ret;

		  ch->pmode = &serial_serocco_ports[cur_asyn_port];
		  serial_serocco_ports[cur_asyn_port].ch = ch;
		  serial_serocco_ports[cur_asyn_port].port.dev = _dev;
		  ret = uart_add_one_port(&serial_serocco_reg, &serial_serocco_ports[cur_asyn_port].port);
		  pr_debug("%s %d registering %d line %d ret %d\n", __FUNCTION__, __LINE__, cur_asyn_port,
			   serial_serocco_ports[cur_asyn_port].port.line, ret);
		  cur_asyn_port++;
	  }
	  break;
	  case SEROCCO_MODE_SYNC:
	  {
		  struct net_device *netdev;
		  struct serocco_sync_port *sp;
		  int ret;

		  netdev = alloc_hdlcdev(ch);
		  if (!netdev) {
			  printk("Cannot allocate hdlcdev!\n");
			  goto no_serocco;
		  }

		  sp = kmalloc(sizeof(struct serocco_sync_port), GFP_KERNEL);
		  if (!sp) {
			  free_netdev(netdev);
			  printk("Cannot allocate sync port strucutre!\n");	
			  goto no_serocco;
		  }
		  memset(sp, 0, sizeof(struct serocco_sync_port));

		  sp->dev = netdev;
		  sp->ch = ch;
		  ch->pmode = sp;

		  serocco_sync_init(netdev);
		  if ((ret = register_hdlc_device(netdev)) < 0) {
			  printk("register hdlc device failed, return value is %d\n", ret);
			  free_netdev(netdev);
			  kfree(sp);
			  goto no_serocco;
		  }

		  pr_debug("%s %d, %p\n", __FUNCTION__, __LINE__, netdev);
		  dev_to_hdlc(netdev)->priv = sp;
	  }
	  break;
	  }
  }

  dev_set_drvdata(_dev, my_serocco);

  return 0;
  
 no_serocco:
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  kfree(my_serocco->ch_a);
  kfree(my_serocco->ch_b);
  kfree(my_serocco);
  /* unallocat memory */

 mem_alloc:
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  
  /* not present, clear structures  */
  return -ENOENT;
  
}

static int serocco_remove(struct device *_dev)
{
  int cur_port = 0;
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);

  for(cur_port = 0; cur_port < 2; cur_port++) {
	  struct serocco_ch *ch;
	  
	  if (!cur_port)
		  ch = my_serocco->ch_a;
	  else
		  ch = my_serocco->ch_b;
		  
	  /* port type specific clean-up */
	  switch(ch->mode) {
	  case SEROCCO_MODE_ASYNC: 
	  {
		  struct serocco_serial_port *sport = (struct serocco_serial_port *) ch->pmode;
		  
		  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
		  if (sport)
			  uart_remove_one_port(&serial_serocco_reg, &sport->port);
	  }
	  break;
	  case SEROCCO_MODE_SYNC: 
	  {
		  struct serocco_sync_port *sp = (struct serocco_sync_port *) ch->pmode;
		  struct net_device *netdev = sp->dev;
      
		  pr_debug("%s %d %p\n", __FUNCTION__, __LINE__, netdev);
		  unregister_hdlc_device(netdev);
		  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
		  free_netdev(netdev);
		  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
		  kfree(sp);
	  }
	  break;
	  }
  }

  if (my_serocco->ch_a->mode == SEROCCO_MODE_ASYNC ||
      my_serocco->ch_b->mode == SEROCCO_MODE_ASYNC) {
	  uart_unregister_driver(&serial_serocco_reg);
  }

  dev_set_drvdata(_dev, NULL);

  /* free irqs */
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  free_irq(my_serocco->irq, my_serocco);  

  /* deallocate memory */
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  kfree(my_serocco->ch_a);
  kfree(my_serocco->ch_b);
  kfree(my_serocco);

  return 0;
}

static struct device_driver serocco_driver = {
        .name           = "serocco",
        .bus            = &platform_bus_type,
        .probe          = serocco_probe,
        .remove         = serocco_remove,

	.suspend	= serocco_suspend,
	.resume		= serocco_resume,
};

int __init serocco_init(void)
{
  int ret;


  pr_debug("%s %d n serials %d: %d %d\n", __FUNCTION__, __LINE__, serial_serocco_reg.nr,
	   serial_serocco_ports[0].port.line,
	   serial_serocco_ports[1].port.line
	  );
  ret = driver_register(&serocco_driver);
  if (ret != 0) {
    printk("cannot register serocco driver\n");
  }

  return ret;
}

void __exit serocco_exit(void)
{
  pr_debug("%s %d\n", __FUNCTION__, __LINE__);
  driver_unregister(&serocco_driver);
}

module_init(serocco_init);
module_exit(serocco_exit);

MODULE_LICENSE("GPL");
