#ifndef _SEROCCO_H_
#define _SEROCCO_H_ 1

#include <linux/config.h>
#include <linux/types.h>
#include <linux/hdlc.h>
#include <linux/skbuff.h>
#include <asm/io.h>

struct serocco_ch;

struct serocco_pin_s {
  u32 line;
  u8 bit_reg;
  u8 bit_mask;
  u8 neg;
  u8 irq_reg;
  u8 irq_mask;
};

struct serocco_chip {
  struct serocco_ch *ch_a;
  struct serocco_ch *ch_b;

  u32 base;
  u32 base_p;

  struct tasklet_struct tasklet; /* tasklet to be called on irq */
  int irq;			/* irq number */

  struct semaphore lock;	/* lock the serocco */
};

struct serocco_ch {
  struct serocco_chip * chip;

  u32 off;

  int mode;
#define SEROCCO_MODE_MASK 0xf
#define SEROCCO_MODE_OFF 0
#define SEROCCO_MODE_ASYNC 1
#define SEROCCO_MODE_SYNC 2
  void * pmode;			/* mode specific data */
  
  struct serocco_pin_s *pins; 	/* pins for this port */

  int running;
};

struct serocco_serial_port {
  struct uart_port port;
  struct serocco_ch *ch;
  int pin_status_in;		/* current pin status */
  int tx_empty;			/* tx fifo empty */
  char *name;
};

struct serocco_sync_port {
  struct serocco_ch *ch;
  struct net_device * dev;

  /* Tx/Rx */
  struct sk_buff *tx_skbuff;
  volatile uint tx_index;
  struct sk_buff *rx_skbuff;
  volatile uint rx_index;

  /* HDLC data */
  sync_serial_settings settings;
  u32 pad __attribute__ ((aligned (4)));

  /* Misc */
  int last_carrier_status;
  int running;

  /* mode dfinition */
  int clock_mode;
  int clock_rate;		/* baud rate */
  int loopback;
#define SSMODE_LOOP_ON 1
#define SSMODE_LOOP_OFF 0
  int encoding;
  int parity;
};

/* ----- SEROCCO-H registers offset ----------------------------------- */

#define SEROCCO_GCMDR			0x00
   #define SEROCCO_SWR			   0x01
#define SEROCCO_GMODE			0x01
   #define SEROCCO_OD_LOW		   0x00
   #define SEROCCO_EDMA			   0x40
   #define SEROCCO_IPC_HIGH                0x30
   #define SEROCCO_IPC_LOW                 0x10
   #define SEROCCO_IPC_OD                  0x00
   #define SEROCCO_OSCPD		   0x08
   #define SEROCCO_DSHP			   0x02
   #define SEROCCO_GIM			   0x01
#define SEROCCO_GSTAR			0x03
   #define SEROCCO_GPI			   0x80
   #define SEROCCO_DMI			   0x40
   #define SEROCCO_ISA2			   0x20
   #define SEROCCO_ISA1			   0x10
   #define SEROCCO_ISA0			   0x08
   #define SEROCCO_ISB2			   0x04
   #define SEROCCO_ISB1			   0x02
   #define SEROCCO_ISB0			   0x01
#define SEROCCO_GPDIRL                  0x04
#define SEROCCO_GPDIRH                  0x05
#define SEROCCO_GPDATL                  0x06
#define SEROCCO_GPDATH                  0x07
#define SEROCCO_GPIML			0x08
#define SEROCCO_GPIMH			0x09
#define SEROCCO_GPISL                   0x0a
#define SEROCCO_GPISH                   0x0b
#define SEROCCO_RFIFOL			0x10
#define SEROCCO_RFIFOH			0x11
   #define SEROCCO_SB_PE                   0x80  
   #define SEROCCO_SB_FE                   0x40  
#define SEROCCO_XFIFOL			SEROCCO_RFIFOL
#define SEROCCO_XFIFOH			SEROCCO_RFIFOH
#define SEROCCO_STARL			0x12
   #define SEROCCO_CEC			   0x10
   #define SEROCCO_XFW			   0x02
#define SEROCCO_STARH			0x13
   #define SEROCCO_RFNE                    0x40
   #define SEROCCO_CD			   0x20
#define SEROCCO_CMDRL			0x14
   #define SEROCCO_XRES			   0x10
   #define SEROCCO_XF			   0x08
   #define SEROCCO_XME			   0x04
#define SEROCCO_CMDRH			0x15
   #define SEROCCO_RMC			   0x80
   #define SEROCCO_RFRD			   0x02
   #define SEROCCO_RRES			   0x01
#define SEROCCO_CCR0L			0x16
   #define SEROCCO_VIS		           0x80
   #define SEROCCO_BCR		           0x20
   #define SEROCCO_TOE                     0x10
   #define SEROCCO_SSEL		           0x08
   #define SEROCCO_CM7                     0x07
#define SEROCCO_CCR0H			0x17
   #define SEROCCO_POWERUP		   0x80
   #define SEROCCO_BUS_NRZ 	           0x00
   #define SEROCCO_BUS_NRZ_TIME2	   0x30
   #define SEROCCO_BUS_NRZI 	           0x20
   #define SEROCCO_BUS_FM0 	           0x40
   #define SEROCCO_BUS_FM1 	           0x50
   #define SEROCCO_BUS_MANCHESTER          0x60
   #define SEROCCO_MOD_ASYNC		   0x03
   #define SEROCCO_MOD_NONE               -1
#define SEROCCO_CCR1L			0x18
   #define SEROCCO_CRL			   0x80
   #define SEROCCO_C32			   0x40
   #define SEROCCO_DIV			   0x04
   #define SEROCCO_ODS			   0x02
#define SEROCCO_CCR1H			0x19
   #define SEROCCO_ICD                     0x40
   #define SEROCCO_RTS                     0x10
   #define SEROCCO_FRTS                    0x8
   #define SEROCCO_FCTS                    0x4
   #define SEROCCO_CAS                     0x2
#define SEROCCO_CCR2L			0x1a
   #define SEROCCO_MDS_MASK		   0xc0
   #define SEROCCO_MDS_AUTOMODE		   0x00
   #define SEROCCO_MDS_ADDR_2		   0x40
   #define SEROCCO_MDS_ADDR_01		   0x80
   #define SEROCCO_MDS_ADDR_EXT		   0xc0
   #define SEROCCO_ADM			   0x20
   #define SEROCCO_PPPM_MASK		   0x0c
   #define SEROCCO_NO_PPP		   0x00
   #define SEROCCO_OCTET_SYNC		   0x04
   #define SEROCCO_BIT_SYNC		   0x0c
#define SEROCCO_CCR2H			0x1b
   #define SEROCCO_ITF                     0x08
#define SEROCCO_CCR3L			0x1c
   #define SEROCCO_RAC			   0x08
   #define SEROCCO_XBRK			   0x02
   #define SEROCCO_ESS7			   0x01
   #define SEROCCO_STOP			   0x01
   #define SEROCCO_CHL_MASK                0x30
   #define SEROCCO_CHL_CS5                 0x30
   #define SEROCCO_CHL_CS6                 0x20
   #define SEROCCO_CHL_CS7                 0x10
   #define SEROCCO_CHL_CS8                 0x00

#define SEROCCO_CCR3H			0x1d
   #define SEROCCO_PAR_MASK                0xc0
   #define SEROCCO_PAR_SPACE               0x00
   #define SEROCCO_PAR_ODD                 0x40
   #define SEROCCO_PAR_EVEN                0x80
   #define SEROCCO_PAR_MARK                0xc0
   #define SEROCCO_PARE                    0x20
   #define SEROCCO_RADD			   0x10
   #define SEROCCO_RFDF                    0x08
   #define SEROCCO_RFTH_MASK               0x03
   #define SEROCCO_RFTH_02                 0x00
   #define SEROCCO_RFTH_04                 0x01
   #define SEROCCO_RFTH_16                 0x02
   #define SEROCCO_RFTH_32                 0x03
#define SEROCCO_TOLEN                   0x1f
   #define SEROCCO_TOIE                    0x80   
#define SEROCCO_BRRL			0x38
#define SEROCCO_BRRH			0x39
#define SEROCCO_RAH1			0x41
#define SEROCCO_RAH2			0x43
#define SEROCCO_RLCRL			0x48
#define SEROCCO_RLCRH			0x49
   #define SEROCCO_RCE			   0x80
#define SEROCCO_ISR0			0x50
   #define SEROCCO_RPF			   0x08
   #define SEROCCO_RME			   0x04
   #define SEROCCO_FLEX			   0x01
#define SEROCCO_IMR0			0x54
/*    #define SEROCCO_RDO			   0x80 */
   #define SEROCCO_RFO			   0x40
   #define SEROCCO_PCE			   0x20
   #define SEROCCO_RSC			   0x10
   #define SEROCCO_RPF			   0x08
   #define SEROCCO_RME			   0x04
   #define SEROCCO_RFS			   0x02
   #define SEROCCO_TIME			   0x02
   #define SEROCCO_FLEX			   0x01
#define SEROCCO_ISR1			0x51
   #define SEROCCO_XDU			   0x04
   #define SEROCCO_BRK			   0x02
#define SEROCCO_IMR1			0x55
   #define SEROCCO_TIN			   0x80
   #define SEROCCO_CSC			   0x40
   #define SEROCCO_XMR			   0x20
   #define SEROCCO_XPR			   0x10
   #define SEROCCO_ALLS			   0x08
   #define SEROCCO_XDU			   0x04
   #define SEROCCO_SUEX			   0x02
   #define SEROCCO_BRK			   0x02
#define SEROCCO_ISR2			0x52
#define SEROCCO_IMR2			0x56
   #define SEROCCO_PLLA			   0x02
   #define SEROCCO_CDSC			   0x01
#define SEROCCO_RSTA			0x58
   #define SEROCCO_VFR			   0x80
   #define SEROCCO_RDO			   0x40
   #define SEROCCO_CRCOK		   0x20
   #define SEROCCO_RAB			   0x10
#define SEROCCO_RBCL			0xc6
#define SEROCCO_RBCH			0xc7
#define SEROCCO_VER0			0xec
#define SEROCCO_VER1			0xed
#define SEROCCO_VER2			0xee
#define SEROCCO_VER3			0xef

#endif
